use std::fmt;
use std::error::Error as StdError;
use std::net::SocketAddr;

use hyper::{self, Method, Uri, Headers, HttpVersion};
use hyper::header::{Header, ContentType};

use {HandleError, Outcome, Response, Responder};
use body::{Body, FromBody};
use storage::Storage;


/// A context during handling.
///
/// It contains an HTTP request, HTTP response to return,
/// and a typemap in order to share variables between middlewares.
#[allow(dead_code)]
pub struct Context {
    version: HttpVersion,
    method: Method,
    uri: Uri,
    headers: Headers,
    remote_addr: Option<SocketAddr>,
    body: Option<Body>,
    shared: Storage,
    states: Storage,
}

impl Context {
    /// Create a new instance of `Context` from components.
    pub fn new(req: hyper::Request, shared: Storage) -> Self {
        // deconstruct hyper::Request
        let remote_addr = req.remote_addr();
        let (method, uri, version, headers, body) = req.deconstruct();

        Context {
            version,
            method,
            uri,
            headers,
            remote_addr,
            body: Some(Body::from(body)),
            shared,
            states: Default::default(),
        }
    }

    /// Return the HTTP version
    pub fn version(&self) -> HttpVersion {
        self.version
    }

    /// Return the reference of request method
    pub fn method(&self) -> &Method {
        &self.method
    }

    /// Return the path of request URL
    pub fn path(&self) -> &str {
        self.uri.path()
    }

    /// Return the query string of request URL
    pub fn query(&self) -> Option<&str> {
        self.uri.query()
    }

    /// Get a reference of header field of `T`.
    pub fn header<T: Header>(&self) -> Option<&T> {
        self.headers.get::<T>()
    }

    /// Return a reference of the request headers
    pub fn headers(&self) -> &Headers {
        &self.headers
    }

    /// Return a mutable reference of the request headers
    pub fn headers_mut(&mut self) -> &mut Headers {
        &mut self.headers
    }

    /// Return a reference of the temporary storage, available during request handling.
    pub fn states(&self) -> &Storage {
        &self.states
    }

    /// Return a mutable reference of the temporary storage, available during request handling.
    pub fn states_mut(&mut self) -> &mut Storage {
        &mut self.states
    }

    /// Return a reference of the global, shared storage.
    pub fn shared(&self) -> &Storage {
        &self.shared
    }

    pub fn take_body<T: FromBody>(&mut self) -> Outcome<T> {
        match self.headers.get::<ContentType>() {
            Some(&ContentType(ref m)) if T::check_mime(m) => (),
            _ => return Outcome::err(HandleError::bad_request()),
        }
        match self.body.take() {
            Some(body) => Outcome::boxed(body.into_value::<T>()),
            None => Outcome::err(HandleError::from_err(EmptyBody)),
        }
    }


    /// Create returned future of `Response` from given `Responder`.
    pub fn respond<R: Responder>(self, responder: R) -> Outcome<Response> {
        Outcome::result(responder.respond(self))
    }
}



#[derive(Debug)]
struct EmptyBody;

impl fmt::Display for EmptyBody {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

impl StdError for EmptyBody {
    fn description(&self) -> &str {
        "empty body"
    }
}
