use std::collections::HashMap;
use std::error::Error;
use std::fmt;
use {StatusCode, Response};
use hyper::header::ContentType;


/// Error type thrown during handling
#[derive(Debug)]
pub struct HandleError {
    status: StatusCode,
    inner: Option<Box<Error + Send>>,
}

impl HandleError {
    pub fn from_status(status: StatusCode) -> Self {
        Self {
            status,
            inner: None,
        }
    }

    pub fn from_err<E: Error + Send + 'static>(err: E) -> Self {
        Self {
            status: StatusCode::InternalServerError,
            inner: Some(Box::new(err)),
        }
    }

    pub fn bad_request() -> Self {
        Self::from_status(StatusCode::BadRequest)
    }

    pub fn unauthorized() -> Self {
        Self::from_status(StatusCode::Unauthorized)
    }

    pub fn not_found() -> Self {
        Self::from_status(StatusCode::NotFound)
    }
}

impl fmt::Display for HandleError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self.inner {
            Some(ref e) => write!(f, "InternalServerError: {}", e),
            None => write!(f, "{:?}", self.status),
        }
    }
}



pub trait ErrorHandler: Send + Sync + 'static {
    fn call(&self, err: HandleError) -> Response;
}

impl<F> ErrorHandler for F
where
    F: Send + Sync + 'static + Fn(HandleError) -> Response,
{
    fn call(&self, err: HandleError) -> Response {
        (*self)(err)
    }
}


pub struct DefaultErrorHandler {
    inner: HashMap<StatusCode, &'static str>,
}

macro_rules! register_template {
    ($map:expr, $code:ident, $description:expr) => {
        $map.insert(StatusCode:: $code, concat!(
r#"<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8" />
  <title>"#, stringify!($code) ,r#"</title>
</head>
<body>
  <h1>"#, $description, r#"</h1>
</body>
</html>"#));
    }
}

impl Default for DefaultErrorHandler {
    fn default() -> Self {
        let mut inner = ::std::collections::HashMap::default();
        // TODO: define more templates
        register_template!(inner, BadRequest, "400 Bad Request");
        register_template!(inner, Unauthorized, "401 Unauthorized");
        register_template!(inner, NotFound, "404 Not Found");
        register_template!(inner, MethodNotAllowed, "405 Method Not Allowed");
        register_template!(inner, InternalServerError, "500 Internal Server Error");
        Self { inner }
    }
}

impl ErrorHandler for DefaultErrorHandler {
    fn call(&self, err: HandleError) -> Response {
        Response::new()
            .with_status(err.status.clone())
            .with_header(ContentType::html())
            .with_body(*self.inner.get(&err.status).unwrap_or(
                &"unknown error",
            ))
    }
}
