use futures::{future, Future, Poll, BoxFuture};
use futures::future::FutureResult;
use HandleError;


/// A future returned from middlewares/handlers.
pub enum Outcome<T, E = HandleError> {
    #[doc(hidden)]
    Result(FutureResult<T, E>),
    #[doc(hidden)]
    Boxed(BoxFuture<T, E>),
}

impl<T, E> Outcome<T, E> {
    /// Create an `Outcome` from a success value
    pub fn ok(val: T) -> Self {
        Outcome::Result(future::ok(val))
    }

    /// Create an `Outcome` from an error value
    pub fn err(err: E) -> Self {
        Outcome::Result(future::err(err))
    }

    /// Create an `Outcome` from a result value
    pub fn result(val: Result<T, E>) -> Self {
        Outcome::Result(future::result(val))
    }

    /// Create an `Outcome` from a future.
    pub fn boxed<F>(f: F) -> Self
    where
        F: Future<Item = T, Error = E> + Send + 'static,
    {
        Outcome::Boxed(f.boxed())
    }
}

impl<T: Send + 'static, E: Send + 'static> Future for Outcome<T, E> {
    type Item = T;
    type Error = E;

    #[inline(always)]
    fn poll(&mut self) -> Poll<Self::Item, Self::Error> {
        match *self {
            Outcome::Result(ref mut f) => f.poll(),
            Outcome::Boxed(ref mut f) => f.poll(),
        }
    }

    #[inline(always)]
    fn boxed(self) -> BoxFuture<Self::Item, Self::Error> {
        match self {
            Outcome::Result(f) => f.boxed(),
            Outcome::Boxed(f) => f,
        }
    }
}

#[macro_export]
macro_rules! try_f {
    ($e:expr) => (match $e {
        Ok(val) => val,
        Err(err) => return ::susanoo::Outcome::err(::susanoo::HandleError::from_err(err)),
    });
}
