use std::error::Error as StdError;
use hyper::{Response as RawResponse, StatusCode};
use hyper::header::ContentType;

use {Context, HandleError};
use super::Response;


/// Represents a "responder", which converted to `hyper::Response`.
pub trait Responder {
    /// Convert itself to `hyper::Response`.
    fn respond(self, ctx: Context) -> Result<Response, HandleError>;
}

impl Responder for Response {
    fn respond(self, _: Context) -> Result<Response, HandleError> {
        Ok(self)
    }
}

impl Responder for RawResponse {
    fn respond(self, _: Context) -> Result<Response, HandleError> {
        Ok(Response::from_raw(self))
    }
}

impl Responder for &'static str {
    fn respond(self, _: Context) -> Result<Response, HandleError> {
        Ok(Response::from_raw(
            RawResponse::new()
                .with_header(ContentType::plaintext())
                .with_body(self),
        ))
    }
}

impl Responder for String {
    fn respond(self, _: Context) -> Result<Response, HandleError> {
        Ok(Response::from_raw(
            RawResponse::new()
                .with_header(ContentType::plaintext())
                .with_body(self),
        ))
    }
}

impl Responder for () {
    fn respond(self, _: Context) -> Result<Response, HandleError> {
        Ok(Response::from_raw(
            RawResponse::new().with_status(StatusCode::NoContent),
        ))
    }
}

impl<T: Responder> Responder for Option<T> {
    fn respond(self, req: Context) -> Result<Response, HandleError> {
        self.ok_or_else(HandleError::not_found).and_then(
            |res| {
                res.respond(req)
            },
        )
    }
}

impl<T, E> Responder for Result<T, E>
where
    T: Responder,
    E: StdError + Send + 'static,
{
    fn respond(self, req: Context) -> Result<Response, HandleError> {
        self.map_err(HandleError::from_err).and_then(
            |res| {
                res.respond(req)
            },
        )
    }
}

impl<T> Responder for Result<T, HandleError>
where
    T: Responder,
{
    fn respond(self, req: Context) -> Result<Response, HandleError> {
        self.and_then(|res| res.respond(req))
    }
}
