use std::borrow::Cow;
use std::sync::Arc;
use hyper::Method;
use regex::Regex;

use handler::Handler;
use super::regex::RegexPattern;


/// A route endpoint, which contains the instance of handler and middlewares
#[allow(dead_code)]
#[derive(Clone)]
pub struct Route {
    pub(super) method: Method,
    pub(crate) pattern: RegexPattern,
    pub(crate) handler: Arc<Handler>,
}

impl Route {
    pub fn new<S, H>(method: Method, pattern: S, handler: H) -> Self
    where
        S: AsRef<str>,
        H: Handler,
    {
        let pattern = normalize_pattern(pattern.as_ref());
        let pattern = Regex::new(&pattern).unwrap();
        Route {
            method: method,
            pattern: pattern.into(),
            handler: Arc::new(handler),
        }
    }

    #[inline(always)]
    pub fn get<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Get, pattern, handler)
    }

    #[inline(always)]
    pub fn post<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Post, pattern, handler)
    }

    #[inline(always)]
    pub fn put<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Put, pattern, handler)
    }

    #[inline(always)]
    pub fn delete<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Delete, pattern, handler)
    }

    #[inline(always)]
    pub fn head<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Head, pattern, handler)
    }

    pub fn pattern_str(&self) -> &str {
        self.pattern.as_str()
    }
}



fn normalize_pattern(pattern: &str) -> Cow<str> {
    let pattern = pattern
        .trim()
        .trim_left_matches("^")
        .trim_right_matches("$")
        .trim_right_matches("/");
    match pattern {
        "" => "^/$".into(),
        s => format!("^{}/?$", s).into(),
    }
}

#[cfg(test)]
mod tests {
    use super::normalize_pattern;

    #[test]
    fn normalize_cases() {
        assert_eq!(normalize_pattern("/"), "^/$");
        assert_eq!(normalize_pattern("/path/to"), "^/path/to/?$");
        assert_eq!(normalize_pattern("/path/to/"), "^/path/to/?$");
    }
}
