use std::any::{Any, TypeId};
use std::collections::HashMap;
use std::hash::{BuildHasherDefault, Hasher};
use std::sync::Arc;


/// A storage of shared, global variables.
#[derive(Clone, Default)]
pub struct Storage {
    inner: Arc<HashMap<TypeId, Box<Any + Send>, BuildHasherDefault<IdentHash>>>,
}

impl Storage {
    /// Try to get the reference of a value typed `T`.
    pub fn try_get<'a, T: Send + Sync + 'static>(&'a self) -> Option<&'a T> {
        let type_id = TypeId::of::<T>();
        self.inner.get(&type_id).and_then(
            |v| v.downcast_ref::<T>(),
        )
    }

    /// Get the reference of a value typed `T`.
    ///
    /// This method will throw a panic if the value of `T` has yet set.
    pub fn get<'a, T: Send + Sync + 'static>(&'a self) -> &'a T {
        self.try_get().expect(
            "Any value of given type has yet set.",
        )
    }

    /// Set a new value of type `T`.
    ///
    /// It is recommended to avoid to call this method after launching the server.
    pub fn set<T: Send + Sync + 'static>(&mut self, value: T) -> bool {
        let type_id = TypeId::of::<T>();
        if self.inner.contains_key(&type_id) {
            return false;
        }

        let inner = Arc::get_mut(&mut self.inner).expect("failed to get a mutable reference");
        inner.insert(type_id, Box::new(value));
        true
    }
}

unsafe impl Send for Storage {}
unsafe impl Sync for Storage {}


struct IdentHash(u64);

impl Hasher for IdentHash {
    fn finish(&self) -> u64 {
        self.0
    }

    fn write(&mut self, bytes: &[u8]) {
        for &byte in bytes {
            self.write_u8(byte);
        }
    }

    fn write_u8(&mut self, byte: u8) {
        self.0 = (self.0 << 8) | (byte as u64);
    }

    fn write_u64(&mut self, bytes: u64) {
        self.0 = bytes;
    }
}

impl Default for IdentHash {
    fn default() -> Self {
        Self { 0: 0 }
    }
}

#[cfg(test)]
mod tests {
    use super::Storage;
    use std::any::TypeId;

    fn get_type_id<T: 'static>(_: &T) -> TypeId {
        TypeId::of::<T>()
    }

    #[test]
    fn case1() {
        let mut storage = Storage::default();
        assert!(storage.set(42u64));
        assert!(storage.set(7i64));
        assert!(!storage.set(90u64));

        assert_eq!(*storage.get::<u64>(), 42u64);
        assert_eq!(get_type_id(storage.get::<u64>()), TypeId::of::<u64>());

        assert_eq!(*storage.get::<i64>(), 7i64);
        assert_eq!(get_type_id(storage.get::<i64>()), TypeId::of::<i64>());
    }
}
