use std::error;
use std::fmt;
use std::io;
use std::net::SocketAddr;
use std::panic::AssertUnwindSafe;
use std::sync::Arc;

use futures::{Future, Poll, Async};
use futures::future::CatchUnwind;
use hyper;
use hyper::server::{Http, Service, NewService};
use tokio_core::reactor::Remote;
use tokio_proto::TcpServer;

use {Context, Outcome, Response, HandleError, Handler, ErrorHandler};
use error::DefaultErrorHandler;
use router::{Router, Route};
use storage::Storage;


/// Root instance of Web application.
#[derive(Default)]
pub struct Susanoo<H: Handler = Router, E: ErrorHandler = DefaultErrorHandler> {
    listen_address: Option<SocketAddr>,
    num_threads: Option<usize>,
    shared: Storage,
    handler: H,
    error_handler: E,
}

impl<H: Handler, E: ErrorHandler> Susanoo<H, E> {
    pub fn with_state<T: Send + Sync + 'static>(mut self, state: T) -> Self {
        self.shared.set(state);
        self
    }

    pub fn with_error_handler(mut self, error_handler: E) -> Self {
        self.error_handler = error_handler;
        self
    }

    /// Set listen address of the application.
    ///
    /// The default address is `"0.0.0.0:4000"`.
    pub fn bind<S: Into<SocketAddr>>(mut self, addr: S) -> Self {
        self.listen_address = Some(addr.into());
        self
    }

    /// Set the number of threads running simulateneous event loops (Unix only).
    ///
    /// The default value is 1.
    #[cfg(not(windows))]
    pub fn threads(mut self, num: usize) -> Self {
        self.num_threads = Some(num);
        self
    }

    /// Start the server with given configuration.
    ///
    /// This method will block the current thread until the server is shut down.
    pub fn run(self) {
        let Self {
            listen_address,
            num_threads,
            handler,
            error_handler,
            shared,
        } = self;

        let listen_address = listen_address.unwrap_or_else(|| "0.0.0.0:4000".parse().unwrap());
        let num_threads = num_threads.unwrap_or(1);

        let handler = Arc::new(handler);
        let error_handler = Arc::new(error_handler);

        let mut server = TcpServer::new(Http::new(), listen_address);
        server.threads(num_threads);
        server.with_handle(move |handle| {
            SusanooService {
                handler: handler.clone(),
                shared: shared.clone(),
                remote: handle.remote().clone(),
                error_handler: error_handler.clone(),
            }
        });
    }
}

impl Susanoo<Router> {
    /// Add a route to the application.
    pub fn with_route(mut self, route: Route) -> Self {
        self.handler.add_route(route);
        self
    }
}



/// An asynchronous function executed by Tokio, associated with a HTTP request.
pub struct SusanooService<H: Handler, E: ErrorHandler> {
    handler: Arc<H>,
    error_handler: Arc<E>,
    shared: Storage,
    remote: Remote,
}

impl<H: Handler, E: ErrorHandler> Clone for SusanooService<H, E> {
    fn clone(&self) -> Self {
        Self {
            handler: self.handler.clone(),
            shared: self.shared.clone(),
            remote: self.remote.clone(),
            error_handler: self.error_handler.clone(),
        }
    }
}

impl<H: Handler, E: ErrorHandler> Service for SusanooService<H, E> {
    type Request = hyper::Request;
    type Response = hyper::Response;
    type Error = hyper::Error;
    type Future = HandleFuture<E>;

    fn call(&self, req: Self::Request) -> Self::Future {
        let ctx = Context::new(req, self.shared.clone());
        let outcome = self.handler.call(ctx);
        HandleFuture {
            outcome: AssertUnwindSafe(outcome).catch_unwind(),
            error_handler: self.error_handler.clone(),
        }
    }
}

impl<H: Handler, E: ErrorHandler> NewService for SusanooService<H, E> {
    type Request = hyper::Request;
    type Response = hyper::Response;
    type Error = hyper::Error;
    type Instance = Self;

    fn new_service(&self) -> io::Result<Self::Instance> {
        Ok(self.clone())
    }
}


pub struct HandleFuture<E: ErrorHandler> {
    outcome: CatchUnwind<AssertUnwindSafe<Outcome<Response>>>,
    error_handler: Arc<E>,
}

impl<E: ErrorHandler> Future for HandleFuture<E> {
    type Item = hyper::Response;
    type Error = hyper::Error;

    fn poll(&mut self) -> Poll<Self::Item, Self::Error> {
        match self.outcome.poll() {
            Ok(Async::NotReady) => Ok(Async::NotReady),
            Ok(Async::Ready(Ok(res))) => Ok(Async::Ready(res.into_raw())),
            Ok(Async::Ready(Err(err))) => Ok(Async::Ready(self.error_handler.call(err).into_raw())),
            Err(_) => Ok(Async::Ready(
                self.error_handler
                    .call(HandleError::from_err(CaughtPanic))
                    .into_raw(),
            )),
        }
    }
}


#[derive(Debug)]
struct CaughtPanic;

impl fmt::Display for CaughtPanic {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "caught panic")
    }
}

impl error::Error for CaughtPanic {
    fn description(&self) -> &str {
        "caught panic"
    }
}
